/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

%{
#include "LVMsyntax.h"
#include "LVM.h"
#include <stdio.h>

extern int yylineno;
extern char* yytext;

extern int current_instruction;

#define return_instruction(X) \
        return current_instruction=X

%}

%option yylineno
%option noyywrap
%option case-insensitive

DIGIT      [0-9]
LETTER     [a-z]|[A-Z]|"_"
SIGN       ("+"|"-")
INTEGER    {SIGN}?{DIGIT}+
CESCAPE    \\.|\\{DIGIT}|\\{DIGIT}{DIGIT}|\\{DIGIT}{DIGIT}{DIGIT}
CSTRING    \"([^\n\\"]|{CESCAPE})*\"
STRING     {CSTRING}
POINTED    {DIGIT}*"."{DIGIT}+|{DIGIT}+"."{DIGIT}*
FLOAT      {SIGN}?{POINTED}|{POINTED}e{SIGN}??{DIGIT}+
COMMENT    "/*"([^/]|"/"[^*])*"*/"|"//"(.)*\n
WHITESPACE [ \t\r\n]+
IDENTIFIER {LETTER}({LETTER}|{DIGIT})*
LABEL      {IDENTIFIER}":"

%%
{COMMENT}    {/* Do nothing */}
{WHITESPACE} {/* Do nothing */}

in           {return_instruction(INPUT_INTEGER);}
in_f         {return_instruction(INPUT_FLOAT);}
in_s         {return_instruction(INPUT_STRING);}

infl         {return_instruction(INTEGER_TO_FLOAT);}
flin         {return_instruction(FLOAT_TO_INTEGER);}

pshf         {return_instruction(PUSH_FRAME);}
popf         {return_instruction(POP_FRAME);}
popfr        {return_instruction(POP_FRAME_AND_RETURN);}
lcl          {return_instruction(ACCESS_LOCAL);}
nlcl         {return_instruction(LOAD_NON_LOCAL_ENVIRONMENT);}

ccat         {return_instruction(CONCATENATE_STRINGS);}
frto         {return_instruction(FROM_TO_STRING);}

and          {return_instruction(AND);}
or           {return_instruction(OR);}
not          {return_instruction(NOT);}

nil          {return_instruction(NIL);}
null         {return_instruction(IS_NULL);}
nnull        {return_instruction(IS_NOT_NULL);}

cr           {return_instruction(CREATE_INTEGER);}
cr_f         {return_instruction(CREATE_FLOAT);}
cr_s         {return_instruction(CREATE_STRING);}
cr_a         {return_instruction(CREATE_ARRAY);}

mka          {return_instruction(MAKE_ARRAY);}

cons         {return_instruction(CONS);}
car          {return_instruction(CAR);}
cdr          {return_instruction(CDR);}
rplaca       {return_instruction(RPLACA);}
rplacd       {return_instruction(RPLACD);}

out          {return_instruction(OUTPUT);}
outl         {return_instruction(OUTPUT_LIST);}
outf         {return_instruction(OUTPUT_FAST);}
outlf        {return_instruction(OUTPUT_LIST_FAST);}

lda          {return_instruction(LOAD_ABSOLUTE);}
sta          {return_instruction(STORE_ABSOLUTE);}
ldc          {return_instruction(LOAD_CONSTANT);}
ld           {return_instruction(LOAD);}
st           {return_instruction(STORE);}
as           {return_instruction(ASSIGN);}
swp          {return_instruction(SWAP);}

pop          {return_instruction(POP);}

cpy          {return_instruction(COPY);}
cln          {return_instruction(CLONE);}

lde          {return_instruction(LOAD_ARRAY_ELEMENT);}
ldei         {return_instruction(LOAD_ARRAY_ELEMENT_IMMEDIATE);}
ste          {return_instruction(STORE_ARRAY_ELEMENT);}

umin         {return_instruction(UNARY_MINUS);}
umin_f       {return_instruction(UNARY_MINUS_FLOAT);}

sz           {return_instruction(SIZE_OF_ARRAY);}
len          {return_instruction(LENGTH_OF_STRING);}
rsz          {return_instruction(RESIZE_ARRAY);}

add          {return_instruction(PLUS);}
sub          {return_instruction(MINUS);}
mul          {return_instruction(TIMES);}
div          {return_instruction(DIVIDED);}
mod          {return_instruction(MODULO);}
pow          {return_instruction(POWER);}

add_f        {return_instruction(PLUS_FLOAT);}
sub_f        {return_instruction(MINUS_FLOAT);}
mul_f        {return_instruction(TIMES_FLOAT);}
div_f        {return_instruction(DIVIDED_FLOAT);}
pow_f        {return_instruction(POWER_FLOAT);}

addi         {return_instruction(PLUS_IMMEDIATE);}
subi         {return_instruction(MINUS_IMMEDIATE);}
muli         {return_instruction(TIMES_IMMEDIATE);}
divi         {return_instruction(DIVIDED_IMMEDIATE);}
modi         {return_instruction(MODULO_IMMEDIATE);}
powi         {return_instruction(POWER_IMMEDIATE);}

lt           {return_instruction(LESS_THAN);}
gt           {return_instruction(GREATER_THAN);}
eq           {return_instruction(EQUAL);}
neq          {return_instruction(NOT_EQUAL);}
lte          {return_instruction(LESS_THAN_OR_EQUAL);}
gte          {return_instruction(GREATER_THAN_OR_EQUAL);}

lt_f         {return_instruction(LESS_THAN_FLOAT);}
gt_f         {return_instruction(GREATER_THAN_FLOAT);}
eq_f         {return_instruction(EQUAL_FLOAT);}
neq_f        {return_instruction(NOT_EQUAL_FLOAT);}
lte_f        {return_instruction(LESS_THAN_OR_EQUAL_FLOAT);}
gte_f        {return_instruction(GREATER_THAN_OR_EQUAL_FLOAT);}

lt_s         {return_instruction(LESS_THAN_STRING);}
gt_s         {return_instruction(GREATER_THAN_STRING);}
eq_s         {return_instruction(EQUAL_STRING);}
neq_s        {return_instruction(NOT_EQUAL_STRING);}
lte_s        {return_instruction(LESS_THAN_OR_EQUAL_STRING);}
gte_s        {return_instruction(GREATER_THAN_OR_EQUAL_STRING);}

lt_g         {return_instruction(LESS_THAN_GENERIC);}
gt_g         {return_instruction(GREATER_THAN_GENERIC);}
eq_g         {return_instruction(EQUAL_GENERIC);}
neq_g        {return_instruction(NOT_EQUAL_GENERIC);}
lte_g        {return_instruction(LESS_THAN_OR_EQUAL_GENERIC);}
gte_g        {return_instruction(GREATER_THAN_OR_EQUAL_GENERIC);}

sm           {return_instruction(ARE_THE_SAME_OBJECT);}
nsm          {return_instruction(ARE_NOT_THE_SAME_OBJECT);}

cls          {return_instruction(CLOSE);}
clsr         {return_instruction(CLOSE_RECURSIVELY);}

call         {return_instruction(CALL_STACK);}
ret          {return_instruction(RETURN);}
retv         {return_instruction(RETURN_VALUE);}

j            {return_instruction(JUMP);}
jz           {return_instruction(JUMP_IF_ZERO);}
jnz          {return_instruction(JUMP_IF_NONZERO);}
nop          {return_instruction(NOP);}
exit         {return_instruction(EXIT);}

{INTEGER}    {return INTEGER_CONSTANT;}
{FLOAT}      {return FLOAT_CONSTANT;}
{STRING}     {return STRING_CONSTANT;}
{IDENTIFIER} {return IDENTIFIER;}
{LABEL}      {return LABEL;}

[^ \t\r\n]+  {return LEXICAL_ERROR;}

%%

int current_instruction;

