/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

%start module

%{
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "symbolstable.h"
#include "LVM.h"
#include "strings.h"
#include "../version.h"

int yylex();

extern int yylineno;
extern char* yytext;

extern FILE *yyin,*yyout;

struct symbols_table* labels;
struct symbols_table* integer_constants;
struct symbols_table* float_constants;
struct symbols_table* string_constants;
struct symbols_table* array_constants;

struct constants_table* constants;

int constants_number=0;

void insert_integer_constant(struct constants_table* t,int x);
void insert_float_constant(struct constants_table* t,double x);
void insert_string_constant(struct constants_table* t,char* x);
void insert_array_constant(struct constants_table* t,int x);


#define output_0                           \
{                                          \
  int i=0;                                 \
  fwrite(&i,sizeof(int),(size_t)1,yyout);  \
}

#define output_integer                     \
{                                          \
  int i;                                   \
  sscanf(yytext,"%i",&i);                  \
  fwrite(&i,sizeof(int),(size_t)1,yyout);  \
}

#define output_float                          \
{                                             \
  double d;                                   \
  sscanf(yytext,"%lf",&d);                    \
  fwrite(&d,sizeof(double),(size_t)1,yyout);  \
}

extern int current_instruction;

void output(int x);
void output_current_instruction();

int instructionsNo=0;
int constantsNo=0;

%}

%token INTEGER_CONSTANT,FLOAT_CONSTANT,STRING_CONSTANT,LABEL
%token IDENTIFIER

%token INPUT_INTEGER,INPUT_FLOAT,INPUT_STRING

%token ARE_THE_SAME_OBJECT,ARE_NOT_THE_SAME_OBJECT

%token ACCESS_LOCAL,LOAD_NON_LOCAL_ENVIRONMENT

%token INTEGER_TO_FLOAT,FLOAT_TO_INTEGER

%token PUSH_FRAME,POP_FRAME,POP_FRAME_AND_RETURN

%token CONCATENATE_STRINGS,FROM_TO_STRING

%token CREATE_INTEGER,CREATE_FLOAT,CREATE_STRING,CREATE_ARRAY,MAKE_ARRAY

%token CONS,CAR,CDR,RPLACA,RPLACD

%token RESIZE_ARRAY,SIZE_OF_ARRAY,LENGTH_OF_STRING

%token NIL,IS_NULL,IS_NOT_NULL
%token OUTPUT,OUTPUT_LIST,OUTPUT_FAST,OUTPUT_LIST_FAST

%token LOAD_CONSTANT
%token LOAD_LABEL /* Not used in symbolic assembly */
%token LOAD_ABSOLUTE,STORE_ABSOLUTE,LOAD,STORE,COPY,CLONE,LOAD_ARRAY_ELEMENT,LOAD_ARRAY_ELEMENT_IMMEDIATE
       STORE_ARRAY_ELEMENT,ASSIGN,POP,SWAP

%token UNARY_MINUS,UNARY_MINUS_FLOAT,PLUS,MINUS,TIMES,DIVIDED,MODULO,POWER
%token PLUS_FLOAT,MINUS_FLOAT,TIMES_FLOAT,DIVIDED_FLOAT,POWER_FLOAT
%token PLUS_IMMEDIATE,MINUS_IMMEDIATE,TIMES_IMMEDIATE,DIVIDED_IMMEDIATE,MODULO_IMMEDIATE,POWER_IMMEDIATE

%token EQUAL,NOT_EQUAL,LESS_THAN,GREATER_THAN,LESS_THAN_OR_EQUAL,
       GREATER_THAN_OR_EQUAL
%token EQUAL_FLOAT,NOT_EQUAL_FLOAT,LESS_THAN_FLOAT,GREATER_THAN_FLOAT,LESS_THAN_OR_EQUAL_FLOAT,
       GREATER_THAN_OR_EQUAL_FLOAT
%token EQUAL_STRING,NOT_EQUAL_STRING,LESS_THAN_STRING,GREATER_THAN_STRING,LESS_THAN_OR_EQUAL_STRING,
       GREATER_THAN_OR_EQUAL_STRING
%token EQUAL_GENERIC,NOT_EQUAL_GENERIC,LESS_THAN_GENERIC,GREATER_THAN_GENERIC,LESS_THAN_OR_EQUAL_GENERIC,
       GREATER_THAN_OR_EQUAL_GENERIC

%token JUMP,JUMP_IF_ZERO,JUMP_IF_NONZERO

%token CALL_STACK,RETURN,RETURN_VALUE

%token CLOSE,CLOSE_RECURSIVELY

%token AND,OR,NOT

%token NOP,EXIT

%token LEXICAL_ERROR
%token END_OF_INSTRUCTIONS /* not a real token */

%%

optional_labels :
  /* nothing */
  |
  LABEL
  {
    int label_code;

    if((label_code=st_lookup(labels,yytext))==-1)
      label_code=st_insert_next(labels,yytext);
    output(LABEL);
    output(label_code);
  }
  optional_labels
  ;

labeled_instruction :
  optional_labels
  instruction      {instructionsNo++;};

nullary_instruction :
  INTEGER_TO_FLOAT|FLOAT_TO_INTEGER|
  CREATE_INTEGER|CREATE_FLOAT|CREATE_STRING|CREATE_ARRAY|
  INPUT_INTEGER|INPUT_FLOAT|INPUT_STRING|
  ARE_THE_SAME_OBJECT|ARE_NOT_THE_SAME_OBJECT|
  CONCATENATE_STRINGS|FROM_TO_STRING|
  NIL|IS_NULL|IS_NOT_NULL|
  OUTPUT|OUTPUT_LIST|OUTPUT_FAST|OUTPUT_LIST_FAST|
  CONS|CAR|CDR|RPLACA|RPLACD|
  COPY|CLONE|LOAD_ARRAY_ELEMENT|STORE_ARRAY_ELEMENT|ASSIGN|POP|
  UNARY_MINUS|UNARY_MINUS_FLOAT|PLUS|MINUS|TIMES|DIVIDED|MODULO|POWER|
  PLUS_FLOAT|MINUS_FLOAT|TIMES_FLOAT|DIVIDED_FLOAT|POWER_FLOAT|
  EQUAL|NOT_EQUAL|LESS_THAN|GREATER_THAN|LESS_THAN_OR_EQUAL|
  GREATER_THAN_OR_EQUAL|
  EQUAL_FLOAT|NOT_EQUAL_FLOAT|LESS_THAN_FLOAT|GREATER_THAN_FLOAT|LESS_THAN_OR_EQUAL_FLOAT|GREATER_THAN_OR_EQUAL_FLOAT|
  EQUAL_STRING|NOT_EQUAL_STRING|LESS_THAN_STRING|GREATER_THAN_STRING|LESS_THAN_OR_EQUAL_STRING|GREATER_THAN_OR_EQUAL_STRING|
  EQUAL_GENERIC|NOT_EQUAL_GENERIC|LESS_THAN_GENERIC|GREATER_THAN_GENERIC|LESS_THAN_OR_EQUAL_GENERIC|
  GREATER_THAN_OR_EQUAL_GENERIC|
  RESIZE_ARRAY|SIZE_OF_ARRAY|LENGTH_OF_STRING|
  RETURN|RETURN_VALUE|
  AND|OR|NOT|
  SWAP|
  POP_FRAME|POP_FRAME_AND_RETURN|
  NOP|EXIT
  ;

unary_instruction_integer :
  LOAD|STORE|LOAD_ABSOLUTE|LOAD_ARRAY_ELEMENT_IMMEDIATE|STORE_ABSOLUTE|CALL_STACK|MAKE_ARRAY|
  PUSH_FRAME|
  ACCESS_LOCAL|LOAD_NON_LOCAL_ENVIRONMENT|
  PLUS_IMMEDIATE|MINUS_IMMEDIATE|TIMES_IMMEDIATE|DIVIDED_IMMEDIATE|MODULO_IMMEDIATE|POWER_IMMEDIATE
  ;

unary_instruction_label :
  JUMP|JUMP_IF_ZERO|JUMP_IF_NONZERO|
  CLOSE|CLOSE_RECURSIVELY
  ;
  
instruction :
  nullary_instruction
  {output_current_instruction();
   output_0;}
  |
  unary_instruction_integer
  {output_current_instruction();}
  INTEGER_CONSTANT
  {output_integer;}
  |
  LOAD_CONSTANT
  LABEL{
    int label_identifier=st_lookup(labels,yytext);

    output(LOAD_LABEL);

    if(label_identifier == -1) /* If the label is unknown then install it: */
      label_identifier=st_insert_next(labels,yytext);

    output(label_identifier);
  }
  |
  LOAD_CONSTANT
  INTEGER_CONSTANT{
    int x;   
 
    output(LOAD_CONSTANT);
    output(constants_number++);
    
    sscanf(yytext,"%i",&x);
    insert_integer_constant(constants,x);
  }
  |
  LOAD_CONSTANT
  FLOAT_CONSTANT{
    double x;

    output(LOAD_CONSTANT);
    output(constants_number++);

    sscanf(yytext,"%lf",&x);
    insert_float_constant(constants,x);
  }
  |
  LOAD_CONSTANT
  STRING_CONSTANT{
    int length=strlen(yytext);
    char* buffer=(char*)malloc(length*sizeof(char)+1);

    output(LOAD_CONSTANT);
    output(constants_number++);

    //yytext[length]='\0';
    fill_buffer_with_yytext(buffer,((char*)yytext));
    insert_string_constant(constants,buffer);
    free(buffer);
  }
  |
  unary_instruction_label
  {output_current_instruction();}
  LABEL
  {
    int label_identifier=st_lookup(labels,yytext);
    
    if(label_identifier == -1) /* If the label is unknown then install it: */
      label_identifier=st_insert_next(labels,yytext);
    
    output(label_identifier);
  }
  ;

labeled_instructions :
  labeled_instruction
  labeled_instructions
  |
  /* nothing */
  ;
  
module :
  labeled_instructions
  ;

%%

int yyerror(const char *s){
  fprintf(stderr,"Line %i: %s in assembly file near \"%s\"\n",yylineno,s,yytext);
  exit(EXIT_FAILURE);
}

void output(int x){
  fwrite(&x,sizeof(int),(size_t)1,yyout);
}

void output_current_instruction(){
  output(current_instruction);
}

struct constants_table_element{
  enum type_ID type;
  void* value;
};

#define BLOCK_SIZE 100

struct constants_table{
  struct constants_table_element* constants;
  size_t elements_no;
  size_t array_size;
};

struct constants_table* create_constants_table(){
  struct constants_table* r=
              (struct constants_table*)malloc(sizeof(struct constants_table));

  r->elements_no=0;
  r->array_size=BLOCK_SIZE;
  r->constants=(struct constants_table_element*)
                 malloc(sizeof(struct constants_table_element)*BLOCK_SIZE);

  return r;
}

void destroy_constants_table(struct constants_table* t){
  int i;

  for(i=0;i< t->elements_no;i++)
    free(t->constants[i].value);

  free(t->constants);
  free(t);
}

void add_constants(struct constants_table* t){
  if(t->array_size){
    t->array_size+=BLOCK_SIZE;
    t->constants=
      realloc(t->constants,
	      sizeof(struct constants_table_element)*(t->array_size));
  }
  t->elements_no++;
}

void insert_integer_constant(struct constants_table* t,int x){
  add_constants(t);

  t->constants[t->elements_no-1].type=integer_type;
  t->constants[t->elements_no-1].value=malloc(sizeof(int));
  *(int*)(t->constants[t->elements_no-1].value)=x;
}

void insert_float_constant(struct constants_table* t,double x){
  add_constants(t);

  t->constants[t->elements_no-1].type=float_type;
  t->constants[t->elements_no-1].value=malloc(sizeof(double));
  *(double*)(t->constants[t->elements_no-1].value)=x;
}

void insert_string_constant(struct constants_table* t,char *x){
  size_t size_of_x=strlen(x);  

  add_constants(t);

  t->constants[t->elements_no-1].type=string_type;
  t->constants[t->elements_no-1].value=malloc(sizeof(char)*(size_of_x-1));
  strncpy((char*)(t->constants[t->elements_no-1].value),x+1,size_of_x-2);
  ((char*)(t->constants[t->elements_no-1].value))[size_of_x-2]='\0';
}

void insert_array_constant(struct constants_table* t,int x){
  add_constants(t);

  t->constants[t->elements_no-1].type=array_type;
  t->constants[t->elements_no-1].value=malloc(sizeof(int));
  *(int*)(t->constants[t->elements_no-1].value)=x;
}

void output_constants(struct constants_table* t){
  int i;

  output(END_OF_INSTRUCTIONS);
  output(END_OF_INSTRUCTIONS);

  for(i=0;i<t->elements_no;i++){
    output(t->constants[i].type);
    switch(t->constants[i].type){
      case integer_type: case array_type:{
        output(*(int*)(t->constants[i].value));        
        break;
      }
      case float_type:{
        fwrite((double*)(t->constants[i].value),sizeof(double),1,stdout);
        break;
      }
      case string_type:{
        output((int)(strlen((char*)(t->constants[i].value))+1));
        printf((char*)(t->constants[i].value));
        break;
      }
      default:{
        fprintf(stderr,"output_constants(): this cannot happen!\n");
        exit(EXIT_FAILURE);
      }
    } /* switch */
  } /* for */
}

int main(int argc, char** argv){
  int r;

  constants=create_constants_table();

  labels=st_create();
  integer_constants=st_create();
  float_constants=st_create();
  string_constants=st_create();
  array_constants=st_create();
  
  printf("LVM bytecode v%s",BYTECODE_VERSION);

  r=yyparse();

  output_constants(constants);

  st_destroy(labels);
  st_destroy(integer_constants);
  st_destroy(float_constants);
  st_destroy(string_constants);
  st_destroy(array_constants);

  return r;
}

