/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "environment.h"
#include <string.h>

#include "types.h"


#define INITIAL_ENVIRONMENT_STACK_SIZE   64
#define INITIAL_ENVIRONMENTS_STACK_SIZE  64
#define INITIAL_BINDINGS_STACK_SIZE      64

struct element{
  char name[IDENTIFIER_LENGTH];
  environment_id slink;
  int pos;
  type_t type;
  int precedence; /* not used for all bindings */
};

struct element *environment_stack;
environment_id next_position;
int current_size;

environment_id *environments_stack;
int environments_next_position;
int environments_current_size;

void init_environment(){
  int i;

  current_size=INITIAL_ENVIRONMENT_STACK_SIZE;
  environment_stack=(struct element*)malloc(sizeof(struct element)*current_size);
  next_position=0;

  //for(i=0;i<current_size;i++){
  //  strcpy((char*)(environment_stack[i].name),"");
  //}

  environments_current_size=INITIAL_ENVIRONMENTS_STACK_SIZE;
  environments_stack=(environment_id*)malloc(sizeof(environments_stack)*environments_current_size);
  environments_next_position=0;

  push_environment(-1);

  create_bindings_stack();
}

environment_id top_environment(){
  if(environments_stack[environments_next_position-1]==-1)
    return -1;
  else
    return environments_stack[environments_next_position-1];
}

environment_id previous_environment(){
  if(environments_next_position<2)
    return -1;
  else
    return environments_stack[environments_next_position-2];
}

environment_id pop_environment(){
  return environments_stack[environments_next_position-- -1];
}

void push_environment(environment_id e){
  if(environments_next_position == environments_current_size){
    int i;

    environments_current_size *= 2;
    environments_stack=(environment_id *)realloc(environments_stack,environments_current_size*sizeof(environment_id));
  }

  environments_stack[environments_next_position++]=e;
}

environment_id bind_with_precedence(environment_id e,
                                    const char* identifier,
                                    int n,
                                    type_t type,
                                    int precedence){
  int r=bind(e,identifier,n,type);
  environment_stack[next_position-1].precedence=precedence;

  return r;
}

environment_id bind(environment_id e,
                    const char* identifier,
                    int n,
                    type_t type){
  if(next_position == current_size){
    int i;

    current_size *= 2;
    environment_stack=(struct element*)realloc(environment_stack,current_size*sizeof(struct element));
  }

  strcpy(environment_stack[next_position].name,identifier);
  environment_stack[next_position].slink=e;
  environment_stack[next_position].pos=n;
  environment_stack[next_position].type=type;

  //fprintf(stderr,"%i: [%s @ %i slink = %i]\n",next_position,identifier,n,e);

  return next_position++;
}

void destroy_environment(){
  free(environment_stack);

  destroy_bindings_stack();
}

/* returns 0 iff succeeds */
int lookup(char* identifier,int *jumps_no, int *n, type_t *t){
  environment_id e=environments_stack[environments_next_position-1];

  *jumps_no=0;
  while(e!=((environment_id) -1 )){
    if(!strcmp(identifier,environment_stack[e].name)){
      *n=environment_stack[e].pos;
      *t=environment_stack[e].type;
      return 0;
    } /* if */
    if(environment_stack[e].pos==0)
      (*jumps_no)++;
    e = environment_stack[e].slink;
  } /* while */
  
  return -1;
}

/* returns 0 iff succeeds */
int lookup_with_precedence(char* identifier,int *jumps_no, int *n, type_t *t, int *precedence){
  environment_id e=environments_stack[environments_next_position-1];

  *jumps_no=0;
  while(e!=((environment_id) -1 )){
    if(!strcmp(identifier,environment_stack[e].name)){
      *n=environment_stack[e].pos;
      *t=environment_stack[e].type;
      *precedence=environment_stack[e].precedence;
      return 0;
    } /* if */
    if(environment_stack[e].pos==0)
      (*jumps_no)++;
    e = environment_stack[e].slink;
  } /* while */

  return -1;
}

int *bindings_stack;
int bindings_next_position;
int bindings_current_size;

void create_bindings_stack(){
  bindings_current_size=INITIAL_BINDINGS_STACK_SIZE;
  bindings_stack=(int*)malloc(sizeof(int)*bindings_current_size);
  bindings_next_position=0;
}

void destroy_bindings_stack(){
  free(bindings_stack);
}

void push_bindings(){
  if(bindings_next_position == bindings_current_size){
    int i;

    bindings_current_size *= 2;
    bindings_stack=(int*)realloc(bindings_stack,bindings_current_size*sizeof(int));
  }

  bindings_stack[bindings_next_position++]=0; /* push an empty bindings set */
}

void pop_bindings(){
  --bindings_next_position;
}

int* current_bindings(){
  //  fprintf(stderr,"[Now pos is %i]\n",bindings_next_position);

  return &(bindings_stack[bindings_next_position-1]);
}

void dump_environment(){
  int i;

  fprintf(stderr,"  i  |    name    pos   slink \n----------------------------------\n");
  for(i=top_environment();i>=0;i--)
    fprintf(stderr,"%3i  |%7s%7i%7i\n",i,environment_stack[i].name,environment_stack[i].pos,environment_stack[i].slink);
  fprintf(stderr,"\n");
}

int next_free_position(environment_id e){
  return environment_stack[e].pos+1;
}


