/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

%{
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "epsilon.h"
#include "types.h"
#include "environment.h"
#include "multibuffer.h"

#include "../version.h"

extern int yylineno;
extern char* yytext;
extern FILE *yyin, *yyout;

int current_if_label=1;
int current_function_label=1;
int current_promise_label=1;

int current_let_label=1;

void output(const char* s);
void output_integer(int x);
void output_output(type_t t);

int precedence;
%}

%union{
  struct{ 
          int label; 
        } labeled_form;
  struct{
          char identifier[IDENTIFIER_LENGTH];
          int label;
        } binding_form;
  struct{
          char identifier[IDENTIFIER_LENGTH];
          int label;
          int order; /* IDENTIFIER, INFIX or POSTFIX_OPERATOR */
          int precedence;
        } binding_form_pos;
  struct{ 
          buffer_t buffer; 
          environment_id environment;
          type_t type;
        } typed_expression;
  struct{
          int bindings_no;
        } multibinding_form;
  type_t type_expression;
}

%left LOWEST_PRECEDENCE

%left <binding_form> POSTFIX_OPERATOR_1

%nonassoc <binding_form> REC
%right <binding_form> LET,LETREC

%left <binding_form> POSTFIX_OPERATOR_2

%right BE
%right IN
%right COLON

%left <binding_form> POSTFIX_OPERATOR_3

%right <labeled_form> LAMBDA
%right DOT 
%left EQUAL,DIFFERENT,LESS,GREATER,LESS_OR_EQUAL,GREATER_OR_EQUAL

%left <binding_form> POSTFIX_OPERATOR_4

%nonassoc INTEGER_CONSTANT, STRING_CONSTANT, FLOAT_CONSTANT
%nonassoc INPUT_INTEGER, INPUT_FLOAT, INPUT_STRING

%left <binding_form> POSTFIX_OPERATOR_5

%left ELSE
%left <labeled_form> IF
%token THEN

%left <binding_form> POSTFIX_OPERATOR_6

%left PLUS,MINUS
%left TIMES,DIVIDED,MODULO /* 'x' and ',' must associate the same way */
%left COMMA /* 'x' and ',' must associate the same way */

%left <binding_form> POSTFIX_OPERATOR_7

%right CONS

%left <binding_form> POSTFIX_OPERATOR_8

%left OPEN_PAR,CLOSE_PAR
%left OPEN_BRACKET,CLOSE_BRACKET

%left <binding_form> POSTFIX_OPERATOR_9

%left <binding_form> IDENTIFIER, INFIX_OPERATOR, POSTFIX_OPERATOR
%left TRUE,FALSE

%left <binding_form> POSTFIX_OPERATOR_10

%left AND,OR
%token DEFINE

%left <binding_form> POSTFIX_OPERATOR_11

%right SEMICOLON
%left EMPTY_LIST

%left <binding_form> POSTFIX_OPERATOR_12

%left CONCATENATE
%token INTEGER,FLOAT,STRING,BOOLEAN,ARRAY,LIST,PROMISE

%left <binding_form> POSTFIX_OPERATOR_13

%right ARROW
%right OF

%left <binding_form> POSTFIX_OPERATOR_14

%left FROM,TO,AT,LENGTH

%left <binding_form> POSTFIX_OPERATOR_15

%nonassoc EMPTY
%left HEAD,TAIL

%left <binding_form> POSTFIX_OPERATOR_16

%nonassoc INTEGER_TO_FLOAT,FLOAT_TO_INTEGER
%left POWER
%left FORCE,DELAY

%left <binding_form> POSTFIX_OPERATOR_17

%left NOT
%left UNARY_MINUS,UNARY_PLUS /* only used for precedence */

%left <binding_form> POSTFIX_OPERATOR_18

%left PRED,SUCC,APPLICATION            /* APPLICATION is only used for precedence */

%left <binding_form> POSTFIX_OPERATOR_19

%right SELECTOR
%nonassoc POSTFIX

%left <binding_form> POSTFIX_OPERATOR_20


%start program

%type <typed_expression> program
%type <typed_expression> expression
%type <type_expression> type

%type <typed_expression> application
%type <typed_expression> list_without_left_bracket
%type <typed_expression> tuple

%type <multibinding_form> bindings
/* %type <multibinding_form> recursive_bindings */
%type <binding_form> identifier_or_operator
%type <binding_form_pos> identifier_declaration
%type <binding_form> postfix_operator
%%

expression:
  INTEGER_TO_FLOAT expression
  {
    if($2.type != integer_type)
      yyerror("integer_to_float should be applied to an integer");
    output("\tinfl\t\t// integer_to_float\n");
    $$.type = float_type;
  }
  |
  FLOAT_TO_INTEGER expression
  {
    if($2.type != float_type)
      yyerror("float_to_integer should be applied to a float");
    output("\tflin\t\t// float_to_integer\n");
    $$.type = integer_type;
  }
  |
  FORCE expression
  {
    if(!is_a_promise_type($2.type))
      yyerror("force should be applied to a promise");

    output("\tcall\t0\t// force\n");

    $$.type = element_type($2.type,0);
  }
  |
  DELAY
  {
    output("\tcls\tPROM_");output_integer(current_promise_label);output(":\t// closure for the promise\n");
    push_buffer(new_buffer());
    output("\n/* Code for promise #");output_integer(current_promise_label);output(": */\n");
    output("PROM_");output_integer(current_promise_label++);output(":\n");
    push_environment(bind(top_environment(),""/* a dummy identifier */,0,type_error /* not used */));
  }
  expression
  {
    pop_environment();
    output("\tretv\t\t// return the value of the promise\n");
    pop_buffer();
    $$.type = new_promise_type($3.type);
  }
  |
  EMPTY expression
  {
    if(!is_a_list_type($2.type))
      yyerror("empty should be applied to lists");
    output("\tnull\t\t// empty\n");
    $$.type = boolean_type;
  }
  |
  INPUT_INTEGER expression
  {
    if($2.type != string_type)
      yyerror("input_integer should have a string parameter");
    output("\toutf\t\t// show the input-request message\n");
    output("\tldc\t\"\\n> \"\t// the prompt...\n");
    output("\toutf\t\t// ...show it\n");
    output("\tin\t\t// input an integer number\n");

    $$.type=integer_type;
  }
  |
  INPUT_FLOAT expression
  {
    if($2.type != string_type)
      yyerror("input_float should have a string parameter");
    output("\toutf\t\t// show the input-request message\n");
    output("\tldc\t\"\\n> \"\t// the prompt...\n");
    output("\toutf\t\t// ...show it\n");
    output("\tin_f\t\t// input a float number\n");

    $$.type=float_type;
  }
  |
  INPUT_STRING expression
  {
    if($2.type != string_type)
      yyerror("input_string should have a string parameter");
    output("\toutf\t\t// show the input-request message\n");
    output("\tldc\t\"\\n> \"\t// the prompt...\n");
    output("\toutf\t\t// ...show it\n");
    output("\tin_s\t\t// input a string\n");

    $$.type=string_type;
  }
  |
  PRED expression
  {
    if($2.type != integer_type)
      yyerror("pred is not applied to an integer");
    output("\tsubi\t1\t// pred\n");
    $$.type = integer_type;
  }
  |
  SUCC expression
  {
    if($2.type != integer_type)
      yyerror("succ is not applied to an integer");
    output("\taddi\t1\t// succ\n");
    $$.type = integer_type;
  }
  |
  LENGTH expression
  {
    if($2.type != string_type)
      yyerror("argument of lenght is not a string");
    output("\tlen\t\t// length\n");
    $$.type=integer_type;
  }
  |
  expression CONCATENATE expression
  {
    if(($1.type!=string_type) || ($3.type!=string_type))
      yyerror("type mismatch of @@ operands");
    $$.type=string_type;

    output("\tccat\t\t// @@\n");
  }
  |
  expression AT expression
  {
    if($1.type != string_type)
      yyerror("first argument of @ operator is not a string");
    if($3.type != integer_type)
      yyerror("second argument of @ operator is not an integer");

    $$.type = string_type;
    output("\tsubi\t1\t// decrement\n");
    output("\tcpy\t\t// make the upper bound\n");
    output("\tfrto\t\t// @\n");
  }
  |
  expression FROM
  { 
    if($1.type != string_type)
      yyerror("first argument of slice operator is not a string");
  }
  expression TO
  {
    if($4.type != integer_type)
      yyerror("second argument of slice operator is not an integer");
    output("\tsubi\t1\t// decrement\n");
  }
  expression
  {
    if($7.type != integer_type)
      yyerror("third argument of slice operator is not an integer");
    output("\tsubi\t1\t// decrement\n");
    output("\tfrto\t\t// from..to operator\n");
    $$.type = string_type;
  }
  |
  expression AND expression
  {
    if(($1.type != boolean_type) || ($3.type != boolean_type))
      yyerror("type mismatch in AND operands");
    $$.type=boolean_type;

    output("\tand\t\t// and\n");
  }
  |
  expression OR expression
  {
    if(($1.type != boolean_type) || ($3.type != boolean_type))
      yyerror("type mismatch in OR operands");
    $$.type=boolean_type;

    output("\tor\t\t// or\n");
  }
  |
  NOT expression
  {
    if($2.type != boolean_type)
      yyerror("argument of NOT is not boolean");
    $$.type=boolean_type;

    output("\tnot\t\t// not\n");
  }
  |
  expression EQUAL expression
  {
    // To do: consider are_types_equal() also for other places
    if(!are_types_equal($1.type,$3.type))
      yyerror("type mismatch in = operands");
    else if($1.type == integer_type)
      output("\teq\t\t// =\n");
    else if($1.type == float_type)
      output("\teq_f\t\t// =\n");
    else if($1.type == string_type)
      output("\teq_s\t\t// =\n");
    else
      output("\teq_g\t\t// =\n");

    $$.type=boolean_type;
  }
  |
  expression DIFFERENT expression
  {
    if(!are_types_equal($1.type,$3.type))
      yyerror("type mismatch in =/= operands");
    else if($1.type == integer_type)
      output("\tneq\t\t// =/=\n");
    else if($1.type == float_type)
      output("\tneq_f\t\t// =/=\n");
    else if($1.type == string_type)
      output("\tneq_s\t\t// =/=\n");
    else
      output("\tneq_g\t\t// =/=\n");

    $$.type=boolean_type;
  }
  |
  expression LESS expression
  {
    if(!are_types_equal($1.type,$3.type))
      yyerror("type mismatch in < operands");
    else if($1.type == integer_type)
      output("\tlt\t\t// <\n");
    else if($1.type == float_type)
      output("\tlt_f\t\t// <\n");
    else if($1.type == string_type)
      output("\tlt_s\t\t// <\n");
    else
      output("\tlt_g\t\t// <\n");

    $$.type=boolean_type;
  }
  |
  expression GREATER expression
  {
    if(!are_types_equal($1.type,$3.type))
      yyerror("type mismatch in > operands");
    else if($1.type == integer_type)
      output("\tgt\t\t// >\n");
    else if($1.type == float_type)
      output("\tgt_f\t\t// >\n");
    else if($1.type == string_type)
      output("\tgt_s\t\t// >\n");
    else
      output("\tgt_g\t\t// >\n");

    $$.type=boolean_type;
  }
  |
  expression LESS_OR_EQUAL expression
  {
    if(!are_types_equal($1.type,$3.type))
      yyerror("type mismatch in <= operands");
    else if($1.type == integer_type)
      output("\tlte\t\t// <=\n");
    else if($1.type == float_type)
      output("\tlte_f\t\t// <=\n");
    else if($1.type == string_type)
      output("\tlte_s\t\t// <=\n");
    else
      output("\tlte_g\t\t// <=\n");

    $$.type=boolean_type;
  }
  |
  expression GREATER_OR_EQUAL expression
  {
    if(!are_types_equal($1.type,$3.type))
      yyerror("type mismatch in >= operands");
    else if($1.type == integer_type)
      output("\tgte\t\t// >=\n");
    else if($1.type == float_type)
      output("\tgte_f\t\t// >=\n");
    else if($1.type == string_type)
      output("\tgte_s\t\t// >=\n");
    else
      output("\tgte_g\t\t// >=\n");

    $$.type=boolean_type;
  }
  |
  TRUE
  {
    output("\tldc\t1\t// true\n");
    $$.type=boolean_type;
  }
  |
  FALSE
  {
    output("\tldc\t0\t// false\n");
    $$.type=boolean_type;
  }
  |
  FLOAT_CONSTANT
  {
    output("\tldc\t");output(yytext);output("\n");
    $$.type=float_type;
  }
  |
  STRING_CONSTANT
  {
    output("\tldc\t");output(yytext);output("\n");
    $$.type=string_type;
  }
  |
  INTEGER_CONSTANT
  {
    output("\tldc\t");output(yytext);output("\n");
    $$.type=integer_type;
  }
  |
  expression PLUS expression
  {
    if($1.type!=$3.type)
      yyerror("type mismatch in + operands");
    else if($1.type == integer_type)
      output("\tadd\t\t// +\n");
    else if($1.type == float_type)
      output("\tadd_f\t\t// +\n");
    else
      yyerror("type mismatch in + operands");

    $$.type=$1.type;
  }
  |
  expression MINUS expression
  {
    if($1.type!=$3.type)
      yyerror("type mismatch in - operands");
    else if($1.type == integer_type)
      output("\tsub\t\t// -\n");
    else if($1.type == float_type)
      output("\tsub_f\t\t// -\n");
    else
      yyerror("type mismatch in - operands");

    $$.type=$1.type;
  }
  |
  expression TIMES expression
  {
    if($1.type!=$3.type)
      yyerror("type mismatch in * operands");
    else if($1.type == integer_type)
      output("\tmul\t\t// *\n");
    else if($1.type == float_type)
      output("\tmul_f\t\t// *\n");
    else
      yyerror("type mismatch in * operands");

    $$.type=$1.type;
  }
  |
  expression DIVIDED expression
  {
    if($1.type!=$3.type)
      yyerror("type mismatch in / operands");
    else if($1.type == integer_type)
      output("\tdiv\t\t// /\n");
    else if($1.type == float_type)
      output("\tdiv_f\t\t// /\n");
    else
      yyerror("type mismatch in / operands");

    $$.type=$1.type;
  }
  |
  expression POWER expression
  {
    if($1.type!=$3.type)
      yyerror("type mismatch in ** operands");
    else if($1.type == integer_type)
      output("\tpow\t\t// **\n");
    else if($1.type == float_type)
      output("\tpow_f\t\t// **\n");
    else
      yyerror("type mismatch in ** operands");

    $$.type=$1.type;
  }
  |
  expression MODULO expression
  {
    if(($1.type != integer_type) || ($3.type != integer_type))
      yyerror("type mismatch");
    $$.type=integer_type;

    output("\tmod\t\t// mod\n");
  }
  |
  IF 
  {
    $1.label=current_if_label++;
    output("/* Start of if...then...else #");output_integer($1.label);output(" */\n");
  }
  expression THEN
  {
    if($3.type != boolean_type)
      yyerror("guard not boolean");

    output("\tjz\tELSE_BRANCH_");output_integer($1.label);output(":\n");
    output("/* then branch #");output_integer($1.label);output(" */\n");
  }
  expression ELSE
  {
    output("\tj\tEND_OF_IF_THEN_ELSE_");output_integer($1.label);output(":\nELSE_BRANCH_");output_integer($1.label);output(":\n");
  }
  expression
  { 
    output("END_OF_IF_THEN_ELSE_");output_integer($1.label);output(":\n");
    output("/* End of if...then...else #");output_integer($1.label);output(" */\n");

    $$.type=unify_types($6.type,$9.type);

    if($$.type == type_error)
      yyerror("then and else branches have incompatible types");
  }
  |
  expression CONS expression
  {
    if(!is_a_list_type($3.type))
      yyerror("The second argument of cons must be a list");
    else if(!do_types_match(element_type($3.type,0),$1.type))
      yyerror("cons should make a homogeneous list");
    output("\tcons\t\t// ::\n");
    $$.type=new_list_type(unify_types($1.type,element_type($3.type,0)));
  }
  |
  HEAD expression
  {
    if(!is_a_list_type($2.type))
      yyerror("head must be applied to a list");
    output("\tcar\t\t// head\n");
    $$.type=element_type($2.type,0);
  }
  |
  TAIL expression
  {
    if(!is_a_list_type($2.type))
      yyerror("tail must be applied to a list");
    output("\tcdr\t\t// tail\n");
    $$.type=$2.type;
  }
  |
  LESS
  {
    output("/* < */\n");
  }
  tuple GREATER
  {
    int n=arity_of_tuple_type($3.type);

    $$.type = $3.type;
    output("\tmka\t");output_integer(n);output("\t// make a tuple of ");output_integer(n);output(" elements \n");
    output("/* > */\n");
  }
  |
  expression SELECTOR INTEGER_CONSTANT
  {
    type_t t = $1.type, t1 = $1.type;
    int n,i,length=arity_of_tuple_type(t);

    if(!is_a_pair_type($1.type))
      yyerror("the selector must be applied to a tuple");

    sscanf(yytext,"%i",&n);
    if(n<1)
      yyerror("the selector integer argument must be greater than 0");
    else if(n > length){
      fprintf(stderr,"n is %i, t is %i\n",n,length);
      yyerror("the selector integer argument exceeds tuple length");
    }
    output("\tldei\t");output_integer(n-1);output("\t// |");output_integer(n);output("\n");

    /* The type is represented as a left-deep binary tree. Length-1 left steps, then if n>1 one step right: */
    for(i=0;i<length-n;i++)
      t=element_type(t,0);
    if(n>1)
      t=element_type(t,1);

    $$.type = t;
  }
  |
  OPEN_BRACKET 
  {
    output("/* [ */\n");
  }
  list_without_left_bracket
  {
    $$.type = $3.type;
    output("/* ] */\n");
  }
  /**/
  |
  OPEN_PAR
  {
    output("/* ( */\n");
  } 
  expression CLOSE_PAR
  { 
    $$.type = $3.type;
    output("/* ) */\n");
  }/**/
  |
  LAMBDA
  {
    push_buffer(new_buffer());
    $1.label=current_function_label;
    output("\n/* Code for unnamed lambda-abstraction #");output_integer(current_function_label);output(": */\n");
    output("FUN_");output_integer(current_function_label++);output(":\n");
  }
  identifier_or_operator
  {
    strncpy($3.identifier,yytext,IDENTIFIER_LENGTH);$3.identifier[IDENTIFIER_LENGTH-1]='\0';
  }
  COLON type 
  {
    push_environment(bind(top_environment(),$3.identifier,0,$6));
  }
  DOT expression
  {
    pop_environment();

    $$.type = new_function_type($6 , $9.type); 
    output("\tretv\t\t// return the value on top\n");
    pop_buffer();
    output("\tcls\tFUN_");output_integer($1.label);output(":\t// make a closure with the current environment\n");
  }
  |
  application %prec APPLICATION
  {
    $$.type = $1.type;
  }
  |
  /* ******************************************************* */
  LET
  {
    $1.label=current_let_label++;
    output("/* let */\n");
    push_bindings();
  }
  bindings 
  {
    pop_bindings();
    output("/* in */\n"); 
  } 
  IN 
  {
    output("\tpshf\t");output_integer($3.bindings_no);output("\t// push frame for let body\n");
  }
  expression
  {
    pop_environment();
    $$.type=$7.type;

    output("\tpopfr\t\t// end of let expression \n");
  }
  /* ******************************************************* */
  |
  REC identifier_declaration
  {
    output("/* rec ");output($2.identifier);output(" */\n");
    strncpy($1.identifier,$2.identifier,IDENTIFIER_LENGTH);$1.identifier[IDENTIFIER_LENGTH-1]='\0';
  }
  COLON type
  {
    if(!is_a_function_type($5))
      yyerror("rec should build a function");

    if($2.order == IDENTIFIER)
      push_environment(bind(top_environment(),$1.identifier,0,$5));
    else if($2.order == POSTFIX_OPERATOR)
      push_environment(bind_with_precedence(top_environment(),$1.identifier,0,new_postfix_operator_type(element_type($5,0),
                                                                                                        element_type($5,1)),
                       $2.precedence));

    output("\tclsr\tFUN_");output_integer(current_function_label);output(":\t// recursively close with current environment\n");
    output("\tpshf\t1\t// create a frame for the recursive definition\n");

    output("/* . */\n");
  }
  DOT expression
  {
    if(!do_types_match($8.type,$5))
      yyerror("declared type does not match the actual type");
    output("\tlcl\t1\t// ");output($1.identifier);output("\n");
    output("\tpopfr\t\t// end of rec expression\n");
    pop_environment();

    $$.type = unify_types($8.type,$5);
  }
  |
  LETREC identifier_declaration
  {
    output("/* letrec ");output($2.identifier);output(" */\n");
    strncpy($1.identifier,$2.identifier,IDENTIFIER_LENGTH);$1.identifier[IDENTIFIER_LENGTH-1]='\0';
  }
  COLON type
  {
    if(!is_a_function_type($5))
      yyerror("letrec should define a function");
  
    if($2.order == IDENTIFIER)
      push_environment(bind(top_environment(),$1.identifier,0,$5));
    else if($2.order == POSTFIX_OPERATOR)
      push_environment(bind_with_precedence(top_environment(),$1.identifier,0,new_postfix_operator_type(element_type($5,0),
                                                                                                        element_type($5,1)),
                                            $2.precedence));
    output("\tclsr\tFUN_");output_integer(current_function_label);output(":\t// recursively close with current environment\n");
    output("\tpshf\t1\t// create a frame for the recursive definition\n");

    output("/* be */\n");
  }
  BE expression 
  {
    if(!do_types_match($8.type,$5))
      yyerror("declared type does not match the actual type");
    output("/* in */\n");
  }
  IN expression
  { 
    output("\tpopfr\t\t// end of recursive definition\n");
    pop_environment(); // The scope of the new indentifer ends here.

    $$.type = $11.type;
  }
  |
  IDENTIFIER
  {
    //////////////////////////////////////
    //fprintf(stderr,"Ho trovato un riferimento all'identificatore %s;\n",yytext);
    //dump_environment();
    //////////////////////////////////////

    int jumps,n;
    type_t t;
    
    if(lookup(yytext,&jumps,&n,&t)!=0)
      yyerror("unbound identifier");
    else if(jumps==0){
      output("\tlcl \t");output_integer(n+1);output("\t// ");output(yytext);output("\n");
    }
    else{
      output("\tnlcl\t");output_integer(jumps);output("\t// access non-local environment for ");output(yytext);output("\n");
      output("\tldei\t");output_integer(n+1);output("\t// ");output(yytext);output("\n");
    }
    $$.type=t;
  }
  |
  PLUS expression %prec UNARY_PLUS
  {
    if((!do_types_match($2.type,integer_type))&&(!do_types_match($2.type,float_type)))
      yyerror("type mismatch in unary plus operand");
    $$.type=integer_type;
    /* unary plus: there is no need to translate it into assembly */
  }
  |
  MINUS expression  %prec UNARY_MINUS
  {
    if(do_types_match($2.type, integer_type))
      output("\tumin\t\t// unary minus\n");
    else if(do_types_match($2.type, float_type))
      output("\tumin_f\t\t// unary minus\n");
    else
      yyerror("type mismatch in unary minus operand");

    $$.type=$2.type;
  }
  |
  expression COLON type
  {
    if(!do_types_match($1.type,$3))
      yyerror("type conflicts with optional declaration");
    else
      $$.type=unify_types($1.type,$3);
  }
  ;

application:
  expression expression %prec APPLICATION
  {
    type_t t,f;

    if(!is_a_function_type($1.type))
      yyerror("only functions can be applied");

    f = new_function_type($2.type,new_generic_type());
    t = unify_types(f,$1.type);
    
    if(t==type_error)
      yyerror("type mismatch in application");

    $$.type=element_type(t,1);

    //fprintf(stderr,"\n\tunify  "); output_type(stderr,f); fprintf(stderr,"\n");
    //fprintf(stderr,"\twith   ");   output_type(stderr,$1.type); fprintf(stderr,"\n");
    //fprintf(stderr,"\tmaking ");   output_type(stderr,t); fprintf(stderr,"\n");

    output("\tcall\t1\t// apply the closure on the stack\n");
  }
  |
  expression postfix_operator %prec APPLICATION
  {
    type_t t;
    type_t operator_type; int jumps_no,pos;

    if(lookup(yytext, &jumps_no, &pos, &operator_type) != 0){
      yyerror("operator is not bound");
    }
    t=unify_types(new_function_type(element_type(operator_type,0),
                                    element_type(operator_type,1)),
                  new_function_type($1.type,new_generic_type()));
    if(t==type_error)
      yyerror("type mismatch in postfix operator application");

    if(jumps_no==0){
      output("\tlcl\t");output_integer(pos+1);output("\n");
    }
    else{
      output("\tnlcl\t");output_integer(jumps_no);output("\n");
      output("\tldei\t");output_integer(pos+1);output("\n");
    }

    $$.type=element_type(t,1);
    output("\tswp\t\t// put the postfix operator under the operand\n\tcall\t1\t// apply the closure on the stack\n");
  }
  ;

bindings:
  identifier_declaration
  {
    strncpy($1.identifier,yytext,IDENTIFIER_LENGTH);$1.identifier[IDENTIFIER_LENGTH-1]='\0';
    output("// ");output(yytext);output(" be... \n");

    if((*current_bindings()) != 0)
      push_environment(previous_environment()); // hide bindings from this same let
    else
      push_environment(top_environment()); // just for being able to always use pop_environment(), later
  }
  BE expression
  {
    pop_environment(); // ok, return to the environment we are building

    if($1.order==IDENTIFIER){
      if((*current_bindings()) == 0)
        push_environment(bind(top_environment(),$1.identifier,(*current_bindings())++,$4.type));
      else
        push_environment(bind(pop_environment(),$1.identifier,(*current_bindings())++,$4.type));
    }
    else if($1.order==POSTFIX_OPERATOR){
      if(!is_a_function_type($4.type))
        yyerror("postfix operators must have function type");

      if((*current_bindings()) == 0)
        push_environment(bind_with_precedence(top_environment(),$1.identifier,(*current_bindings())++,
                                              new_postfix_operator_type(element_type($4.type,0),element_type($4.type,1)),
                                              $1.precedence));
      else
        push_environment(bind_with_precedence(pop_environment(),$1.identifier,(*current_bindings())++,
                                              new_postfix_operator_type(element_type($4.type,0),element_type($4.type,1)),
                                              $1.precedence));
    }
    $$.bindings_no=1;
  }
  |
  bindings COMMA bindings
  {
    $$.bindings_no = $1.bindings_no + $3.bindings_no;
  }
  ;

postfix_operator:
  POSTFIX_OPERATOR_1 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_2 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_3 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_4 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_5 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_6 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_7 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_8 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_9 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_10 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_11 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_12 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_13 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_14 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_15 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_16 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_17 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_18 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_19 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  |
  POSTFIX_OPERATOR_20 {strcpy($$.identifier, yytext); $$.label = $1.label;}
  ;

identifier_or_operator:
  IDENTIFIER
  {
    strcpy($$.identifier, yytext);
    $$.label = $1.label;
  }
  |
  postfix_operator
  {
    strcpy($$.identifier, yytext);
    $$.label = $1.label;
  }
  |
  INFIX_OPERATOR
  {
    strcpy($$.identifier, $1.identifier);
    $$.label = $1.label;
  }
  ;

identifier_declaration:
  identifier_or_operator
  {
    strcpy($$.identifier, $1.identifier);
    $$.label = $1.label;
    $$.order=IDENTIFIER;
  }
  |
  POSTFIX INTEGER_CONSTANT
  {
    sscanf(yytext,"%i",&precedence);

    if((precedence<1)||(precedence>20))
      yyerror("operator precedence must be greater than 0 and less than 21");
  }
  identifier_or_operator
  {
    strcpy($$.identifier, $4.identifier);
    $$.label = $4.label;
    $$.order=POSTFIX_OPERATOR;
    $$.precedence=precedence;
  }
  ;

tuple:
  expression COMMA expression
  {
    if(is_a_pair_type($1.type)||is_a_pair_type($3.type))
      yyerror("cannot make a tuple of tuples");

    $$.type = new_pair_type($1.type,$3.type); 
  }
  |
  tuple COMMA expression
  {
    if(is_a_pair_type($3.type))
      yyerror("cannot make a tuple of tuples");

    $$.type = new_pair_type($1.type,$3.type);
  }
  ;

list_without_left_bracket:
  CLOSE_BRACKET
  {
    output("\tnil\t\t// empty list\n");
    $$.type=new_list_type(new_generic_type());
  }
  |
  expression CLOSE_BRACKET
  {
    output("\tnil\t\t// trailing empty list\n\tcons\t\t// ;\n");
    $$.type=new_list_type($1.type);
  }
  |
  expression SEMICOLON list_without_left_bracket
  {
    if(!is_a_list_type($3.type))
      yyerror("The [...;...] notation should build a homogeneous list");
    else if(!do_types_match(element_type($3.type,0),$1.type))
      yyerror("The [...;...] notation should build a homogeneous list");
    output("\tcons\t\t// ;\n");
    $$.type=new_list_type(unify_types($1.type,element_type($3.type,0)));
  }
  ;

type:
  INTEGER { $$ = integer_type; }
  |
  FLOAT   { $$ = float_type; }
  |
  BOOLEAN { $$ = boolean_type; }
  |
  STRING  { $$ = string_type; }
  |
  LIST OF type { $$ = new_list_type($3); }
  |
  ARRAY OF type { $$ = new_array_type($3); }
  |
  PROMISE OF type { $$ = new_promise_type($3); }
  |
  type ARROW type { $$ = new_function_type($1,$3); }
  |
  type TIMES type { $$ = new_pair_type($1,$3); }
  |
  OPEN_PAR type CLOSE_PAR { $$ = $2; }
  |
  identifier_or_operator { $$ = generic_type(yytext); } /*IDENTIFIER !!!!!!*/
  ;

naming:
  DEFINE IDENTIFIER
  {
    output(yytext);output(":\n");
  }
  EQUAL /* To do */
  expression
  {
    /* To do */
  }
  ;

namings:
  naming
  SEMICOLON
  namings
  |
  /* nothing */
  ;

program:
  namings
  {
    output("main:\n");
  }
  expression
  {
    output("\n\n/* Output the result: */\n");
    output_output($3.type);
    output("\n/* Exit with success: */\n\tldc\t0\t// success\n\texit\t\t// exit the LVM\n");
    //output("\n\tldc\t\"\\n\"\n\tout\n\tj\tmain:\t// loop\n");
    fprintf(stderr,"This expression has type ");
    output_type(stderr,$3.type);
    fprintf(stderr,"\n");
  }
  ;

%%

int already_output_boolean=0;
int last_output_list_label=0;

void output_output(type_t t){
  if(is_a_generic_type(t)){
    output("\t/* This should not happen: */\n");
    output("\tpop\tldc\t\"<#generic???>\"\n\tout\n");
  }
  else if(is_a_function_type(t)){
    output("/* Output a function object: */\n\tpop\n");
    output("\tldc\t\"<function>\"\n");
    output("\tout\n");
  }
  else if(is_a_promise_type(t)){
    output("/* Output a function object: */\n\tpop\n");
    output("\tldc\t\"<promise>\"\n");
    output("\tout\n");
  }
  else if(is_a_pair_type(t)){
    int label=++last_output_list_label;
    int i,length=arity_of_tuple_type(t);

    output("/* Output a tuple object: */\n");
    output("\tldc\t\"<\"\n\tout\n");

    for(i=1;i<=length;i++){
      output("\tcpy\t\t// keep another reference to the tuple\n");
      output("/* Output element #");output_integer(i);output("*/\n");
      output("\tldc\t");output_integer(i-1);output("\n");
      output("\tlde\t\t// access the element\n");
      
      output_output(element_of_tuple_type(t,i));
      if(i+1<=length)
        output("\tldc\t\", \"\n\tout\n");
    }
    output("\tpop\t\t// we don't need the tuple anymore.\n");

    output("\tldc\t\">\"\n\tout\n");
  }
  else if(is_a_list_type(t)){
    int label=++last_output_list_label;

    output("/* Output a list object: */\n");
    output("\tldc\t\"[ \"\n\tout\n");

    output("OUTPUT_LIST_");output_integer(label);output(":\n");
    output("\tcpy\n");
    output("\tnull\n");
    output("\tjnz\tOUTPUT_LIST_OUT_");output_integer(label);output(":\n");
    output("\tcpy\n\tcar\n");
    output_output(element_type(t,0));
    output("\tldc\t\"; \"\n\tout\n");
    output("\tcdr\n");
    output("\tj\tOUTPUT_LIST_");output_integer(label);output(":\n");
    output("OUTPUT_LIST_OUT_");output_integer(label);output(":\n");
    output("\tldc\t\"]\"\n\tout\n");
    output("\tpop\n");
  }
  else if(t==boolean_type){
    output("/* Output a boolean object: */\n");
    output("\tldc\tOUTPUT_BOOLEAN:\n");
    output("\tld\t0\t// push current environment\n");
    output("\tcons\t\t// make a closure\n");
    output("\tswp\t\t// swap argument and closure\n");
    output("\tcall\t1\n");
    if(!already_output_boolean){
      push_buffer(new_buffer());
      output("\nOUTPUT_BOOLEAN:\n");
      output("\tlcl\t1\n");
      output("\tjnz\tOUTPUT_TRUE:\n");
      output("\tldc\t\"false\"\n");
      output("\tj\tEND_OF_OUTPUT_BOOLEAN:\n");
      output("OUTPUT_TRUE:\n");
      output("\tldc\t\"true\"\n");
      output("END_OF_OUTPUT_BOOLEAN:\n");
      output("\tout\n");
      output("\tret\n\n");
      pop_buffer();    
      already_output_boolean=1;
    }
  }
  else if(t==string_type){
    output("/* Output a string object: */\n");
    output("\tldc\t\"\\\"\"\n");
    output("\tout\n");
    output("\tout\n");
    output("\tldc\t\"\\\"\"\n");
    output("\tout\n");
  }
  else
    output("\toutf\t\t// output a primitive object\n");
}

int yyerror(char* message){
  fprintf(stderr,"Line %i: %s near \'%s\'\n",yylineno,message,yytext);
  exit(EXIT_FAILURE);
}

void output_integer(int x){
  char s[10];
  sprintf(s,"%i",x);
  output(s);
}

void output(const char* s){
  write_to_buffer(top_buffer(),s);
}

int main(){
  if(isatty(fileno(stdin)))
    fprintf(stderr,
"-------------------------------------------------------------------------------\n"
"                    eeee               i  ll                                   \n"
"                   ee    p ppp   sss       l   ooo  n nnn                      \n"
"                    eee   p   p s     ii   l  o   o  nn  n                     \n"
"                   ee     p   p  sss   i   l  o   o  n   n                     \n"
"                    eeee  pppp      s iii lll  ooo   n   n                     \n"
"                          p      sss                                           \n"
"                         ppp           http://savannah.gnu.org/projects/epsilon\n"
"-------------------------------------------------------------------------------\n"
"epsilon %s, %s comes with ABSOLUTELY NO WARRANTY.\n"
"\n"
"This is free software, and you are welcome to redistribute it under certain\n"
"conditions; see the file COPYING for details.\n"
"\n",
    VERSION_STRING,COPYRIGHT_STRING);

  initialize_types_table();
  init_multibuffer();

  init_environment();

  push_buffer(new_buffer());
  yyparse();

  destroy_environment();

  flush_multibuffer(yyout);
  free_multibuffer();
  destroy_types_table();

  return EXIT_SUCCESS;
}

