/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "types.h"
#include "term.h"

type_t type_error,integer_type,float_type,boolean_type,string_type;
type_t generic_list_type,generic_pair_type,generic_function_type,generic_array_type,generic_promise_type,
       generic_postfix_operator_type,generic_infix_operator_type;

constructor_t list_c,pair_c,function_c,array_c,promise_c,postfix_operator_c,infix_operator_c;

void initialize_types_table(){
  term_t v1,v2,v3;

  init_terms();

  type_error=error;

  list_c=constructor("list",1);
  pair_c=constructor("tuple",2);
  function_c=constructor("function",2);
  array_c=constructor("array",1);
  promise_c=constructor("promise",1);
  postfix_operator_c=constructor("postfix",2);
  infix_operator_c=constructor("infix",3);

  integer_type=term(constructor("integer",0));
  float_type=term(constructor("float",0));
  boolean_type=term(constructor("boolean",0));
  string_type=term(constructor("string",0));

  v1=vterm(variable("."));v2=vterm(variable(".."));v3=vterm(variable("..."));

  generic_list_type=term(list_c,v1);
  generic_pair_type=term(pair_c,v1,v2);
  generic_function_type=term(function_c,v1,v2);
  generic_array_type=term(array_c,v1);
  generic_promise_type=term(promise_c,v1);
  generic_postfix_operator_type=term(postfix_operator_c,v1,v2);
  generic_infix_operator_type=term(infix_operator_c,v1,v2,v3);
}

void destroy_types_table(){
  free_terms();
}

type_t unify_types(type_t a,type_t b){
  return unify(a,b);
}

int do_types_match(type_t a,type_t b){
  return unify(a,b) != type_error;
}

type_t new_list_type(type_t x){
  return term(list_c,x);
}

type_t new_pair_type(type_t x,type_t y){
  return term(pair_c,x,y);
}

type_t new_function_type(type_t x,type_t y){
  return term(function_c,x,y);
}

type_t new_array_type(type_t x){
  return term(array_c,x);
}

type_t new_promise_type(type_t x){
  return term(promise_c,x);
}

type_t new_postfix_operator_type(type_t from, type_t to){
  return term(postfix_operator_c, from, to);
}

type_t new_infix_operator_type(type_t from_1, type_t from_2, type_t to){
  return term(infix_operator_c, from_1, from_2, to);
}

type_t new_generic_type(){
  return vterm(new_variable("t"));
}

type_t generic_type(char* name){
  return vterm(variable(name));
}

int is_a_list_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_list_type);
}

int is_a_pair_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_pair_type);
}

int is_a_function_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_function_type);
}

int is_an_array_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_array_type);
}

int is_a_promise_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_promise_type);
}

int is_a_postfix_operator_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_postfix_operator_type);  
}

int is_an_infix_operator_type(type_t a){
  return (!is_a_generic_type(a)) && do_types_match(a,generic_infix_operator_type);
}

int is_a_generic_type(type_t a){
  return is_variable(a);
}

type_t element_type(type_t t,int n){
  return subterm(t,n);
}

int arity_of_tuple_type(type_t t){
  int r;
  for(r=1;is_a_pair_type(t);r++)
    t=element_type(t,0);

  return r;
}

/* This is O(dimension). n starts at 1. */
type_t element_of_tuple_type(type_t t,int n){
  int length=arity_of_tuple_type(t);
  int i;

  for(i=0;i<length-n;i++)
    t=element_type(t,0);
  if(n>1)
    t=element_type(t,1);

  return t;
}

int are_types_equal(type_t x,type_t y){
  return are_terms_equal(x,y);
}

int are_types_equal_excluding_variables(type_t x,type_t y){
  return are_terms_equal_excluding_variables(x,y);
}

void output_type(FILE* f,type_t t){
  output_term(f,t);
  /*/
  if(is_a_generic_type(t)){
    output_term(f,t);;
  }
  else if(is_a_function_type(t)){
    output_type(f,element_type(t,0));
    fprintf(f," -> ");
    output_type(f,element_type(t,1));
  }
  else if(is_a_pair_type(t)){
    output_type(f,element_type(t,0));
    fprintf(f," * ");
    output_type(f,element_type(t,1));
  }
  else if(is_an_array_type(t)){
    fprintf(f,"array of ");
    output_type(f,element_type(t,0));
  }
  else if(is_a_list_type(t)){
    fprintf(f,"list of ");
    output_type(f,element_type(t,0));
  }
  else if(is_a_promise_type(t)){
    fprintf(f,"promise of ");
    output_type(f,element_type(t,0));
  }
  else
    output_term(f,t);
  /**/
}


